@@ -1,5 +1,9 @@
 """
-Library to control MagicSwitchbot devices (chinese clone of SwitchBot)
+Library to control MagicSwitchbot devices
+
+@author: ec-blaster
+@since: March 2021
+@license: MIT 
 """
 import time
 import binascii
@@ -8,22 +12,74 @@ import random
 from bluepy import btle
 from Crypto.Cipher import AES
 from threading import Timer
+from homeassistant.components.websocket_api.error import Disconnect
+
+'''How many times we will retry in case of error'''
+DEFAULT_RETRY_COUNT = 3
+
+'''How many seconds to wait between retries'''
+DEFAULT_RETRY_TIMEOUT = 1
+
+'''Max seconds to wait before the connection is established''' 
+DEFAULT_CONNECT_TIMEOUT = 3
+
+'''Max seconds to wait before the device sends back the response to a command'''
+NOTIFY_TIMEOUT = 5
 
-DEFAULT_RETRY_COUNT = 0
-DEFAULT_RETRY_TIMEOUT = 0.2
-NOTIFICATION_TIMEOUT = 5
 NO_TIMEOUT = -1
 
 _LOGGER = logging.getLogger(__name__)
 
 
+class PeripheralExt(btle.Peripheral):
+    """Derived class from bluepy
+    
+    This subclass constructor is based  based on Pull Request #374 by Gunnarl,
+    https://github.com/IanHarvey/bluepy/pull/374)
+    It allows us to stablish a connect timeout to the bluetooth device.
+    """
+
+    def __init__(self, deviceAddr=None, addrType=btle.ADDR_TYPE_PUBLIC, iface=None, timeout=None):
+        btle.BluepyHelper.__init__(self)
+        self._serviceMap = None  # Indexed by UUID
+        (self.deviceAddr, self.addrType, self.iface) = (None, None, None)
+
+        if isinstance(deviceAddr, btle.ScanEntry):
+            self._connect(deviceAddr.addr, deviceAddr.addrType, deviceAddr.iface, timeout)
+        elif deviceAddr is not None:
+            self._connect(deviceAddr, addrType, iface, timeout)
+            
+    def _connect(self, addr, addrType=btle.ADDR_TYPE_PUBLIC, iface=None, timeout=None):
+        if len(addr.split(":")) != 6:
+            raise ValueError("Expected MAC address, got %s" % repr(addr))
+        if addrType not in (btle.ADDR_TYPE_PUBLIC, btle.ADDR_TYPE_RANDOM):
+            raise ValueError("Expected address type public or random, got {}".format(addrType))
+        self._startHelper(iface)
+        self.addr = addr
+        self.addrType = addrType
+        self.iface = iface
+        if iface is not None:
+            self._writeCmd("conn %s %s %s\n" % (addr, addrType, "hci" + str(iface)))
+        else:
+            self._writeCmd("conn %s %s\n" % (addr, addrType))
+        rsp = self._getResp('stat', timeout)
+        if rsp is None:
+            raise btle.BTLEDisconnectError("Timed out while trying to connect to peripheral %s, addr type: %s" % 
+                                      (addr, addrType), rsp)
+        while rsp['state'][0] == 'tryconn':
+            rsp = self._getResp('stat', timeout)
+        if rsp['state'][0] != 'conn':
+            self._stopHelper()
+            raise btle.BTLEDisconnectError("Failed to connect to peripheral %s, addr type: %s" % (addr, addrType), rsp)
+
+
 class MagicSwitchbotDelegate (btle.DefaultDelegate):
     """MagicSwitchbotDelegate
     
     Bluetooth notifications subscription to handle responses from the MagicSwitchbot device
     """
 
-    def __init__(self, readHandle):
+    def __init__(self, readHandle, mac):
         """Class constructor
         
         This constructor must receive the handle of the characteristic we want to subscribe to.
@@ -31,19 +87,22 @@ class MagicSwitchbotDelegate (btle.DefaultDelegate):
         ----------
         readHandle : int
             Handle to the bluetooth characteristic we want to subscribe
+        mac : string
+            MAC address of the device
         """
         btle.DefaultDelegate.__init__(self)
         self._readHandle = readHandle
+        self._mac = mac
         self._data = None
         self._received = False
-        _LOGGER.debug("Notification handler for MagicSwtichbot initialized")
+        _LOGGER.debug("MagicSwitchbot[%s] Notification handler initialized", mac)
       
     def resetData(self):
         """Resets the received data through the characteristic
         """
         self._data = None
         self._received = False
-        _LOGGER.debug("Resetting received data")
+        _LOGGER.debug("MagicSwitchbot[%s] Resetting received data", self._mac)
       
     def hasData(self) -> bool:
         """Check if data received
@@ -86,12 +145,11 @@ class MagicSwitchbotDelegate (btle.DefaultDelegate):
         """
         if (cHandle == self._readHandle):
             '''Filter our notified data'''
-            _LOGGER.debug("Received data from device: %s", data)
             self._received = True
             self._data = data
         else:
             '''Discard all other notifications'''
-            _LOGGER.debug("Received data from device at unexpected handle %d: %s", cHandle, data)
+            _LOGGER.debug("MagicSwitchbot[%s] Received data from device at unexpected handle %d: %s", self._mac, cHandle, data)
             btle.DefaultDelegate.handleNotification(self, cHandle, data)
 
 
@@ -116,6 +174,16 @@ class MagicSwitchbotDevice:
     CMD_TIMEDSWITCH = "0508"
     CMD_GETTOKEN = "0601"
     
+    COMMANDS = {
+        CMD_ENNOTIF: "CMD_ENNOTIF",
+        CMD_GETBAT: "CMD_GETBAT",
+        CMD_OTA: "CMD_OTA",
+        CMD_SWITCH: "CMD_SWITCH",
+        CMD_MODIFYPWD: "CMD_MODIFYPWD",
+        CMD_TIMEDSWITCH: "CMD_TIMEDSWITCH",
+        CMD_GETTOKEN: "CMD_GETTOKEN"
+    }
+    
     """Protocol parameters definition"""
     PAR_SWITCHON = "01"
     PAR_SWITCHOFF = "00"
@@ -134,7 +202,7 @@ class MagicSwitchbotDevice:
     STA_OK = "00"
     STA_ERR = "01"
     
-    def __init__(self, mac, retry_count=DEFAULT_RETRY_COUNT, password=None, interface=None) -> None:
+    def __init__(self, mac, retry_count=DEFAULT_RETRY_COUNT, password=None, interface=0, connect_timeout=DEFAULT_CONNECT_TIMEOUT, disconnect_timeout=NO_TIMEOUT) -> None:
         """Creates a new instance to control the device
         
         Parameters
@@ -147,6 +215,8 @@ class MagicSwitchbotDevice:
                 Password or PIN set on the device
             interface : int
                 Number of the bluetooth client interface to use. It will be prefixed by 'hci'. Default: hci0
+            connect_timeout : int
+                Timeout in seconds for every connection. Default: 3 seconds
         
         """
         self._interface = interface
@@ -161,19 +231,25 @@ class MagicSwitchbotDevice:
         self._token = None
         self._battery = None
         self._delegate = None
+        self._connectTimeout = connect_timeout
+        self._disconnectTimeout = disconnect_timeout
+        self._timer = None
         
     def __del__(self):
         self._disconnect()
 
-    def _connect(self, timeout=NO_TIMEOUT) -> bool:
+    def _connect(self, connect_timeout=DEFAULT_CONNECT_TIMEOUT, disconnect_timeout=NO_TIMEOUT, retries=1) -> bool:
         """Connects to the device
         
         This method allows us to connect to the Magic Switchbot device
         
         Params
         ------
-            timeout : int
-                Specifies the ammount of time (seconds) that will be scheduled to automatically
+            connect_timeout : int (Optional)
+                Specifies the amount of time (seconds) we'll be waiting for the bluetooth device
+                to connect. If it doesn't connect on time, it returns False
+            disconnect_timeout : int (Optional)
+                Specifies the amount of time (seconds) that will be scheduled to automatically
                 disconnect from the device. If it's not specified, the client does not disconnect
                 until the object is disposed from memory
         Returns
@@ -181,42 +257,68 @@ class MagicSwitchbotDevice:
             bool
                 Returns True on successful connection
         """
+        
+        """Don't try to reconnect if we are connected already"""
         if self._is_connected():
             return True
         
-        ok = True
-        try:
-            _LOGGER.debug("Connecting to MagicSwitchbot at address %s from hci%d...", self._mac, self._interface)
-            self._device = btle.Peripheral(self._mac,
-                                                  btle.ADDR_TYPE_PUBLIC,
-                                                  self._interface)
-            _LOGGER.info("Connected to MagicSwitchbot at %s from hci%d", self._mac, self._interface)
+        connected = False
+        if disconnect_timeout != NO_TIMEOUT:
+            self._disconnectTimeout = disconnect_timeout
+        partial_connect = False
+        for i in range(retries): 
+            try:
+                _LOGGER.debug("MagicSwitchbot[%s] Connecting using hci%d with %d seconds timeout (%d of %d retries)...", self._mac, self._interface, connect_timeout, (i + 1), retries)
+                self._device = PeripheralExt(deviceAddr=self._mac,
+                                             addrType=btle.ADDR_TYPE_PUBLIC,
+                                             iface=self._interface,
+                                             timeout=connect_timeout)
+                _LOGGER.info("MagicSwitchbot[%s] Connected with hci%d.", self._mac, self._interface)
+                partial_connect = True
+                
+                '''Initialize service and characteristics handles to the device'''
+                self._service = self._device.getServiceByUUID(self.UUID_SERVICE)
+                self._userReadChar = self._service.getCharacteristics(self.UUID_USERREAD_CHAR)[0]
+                self._cccdDescriptor = self._userReadChar.getDescriptors(forUUID=self.UUID_NOTIFY_SET)[0]
+                self._userWriteChar = self._service.getCharacteristics(self.UUID_USERWRITE_CHAR)[0]
+    
+                '''Once we connected, let's enable the response notifications'''
+                self._enableNotifications()
+                
+                connected = True
+                
+                self._scheduleDisconnection()
+            except btle.BTLEDisconnectError as e:
+                if partial_connect:
+                    _LOGGER.error("MagicSwitchbot[%s] Incomplete connection to device (%s)", self._mac, str(e))
+                else:
+                    _LOGGER.error("MagicSwitchbot[%s] Couldn't connect to device (%s)", self._mac, str(e))
+                self._device = None
+            except btle.BTLEException:
+                _LOGGER.error("MagicSwitchbot[%s] Failed to connect to device", self._mac, exc_info=True)
+                self._device = None
+            except Exception:
+                _LOGGER.error("MagicSwitchbot[%s] Error getting device info", self._mac, exc_info=True)
+                self._device = None
             
-            '''Initialize service and characteristics handles to the device'''
-            self._service = self._device.getServiceByUUID(self.UUID_SERVICE)
-            self._userReadChar = self._service.getCharacteristics(self.UUID_USERREAD_CHAR)[0]
-            self._cccdDescriptor = self._userReadChar.getDescriptors(forUUID=self.UUID_NOTIFY_SET)[0]
-            self._userWriteChar = self._service.getCharacteristics(self.UUID_USERWRITE_CHAR)[0]
-
-            '''Once we connected, let's enable the response notifications'''
-            self._enableNotifications()
+            if connected:
+                return True
+            elif i < retries - 1:
+                _LOGGER.debug("MagicSwitchbot[%s] Retrying in %d seconds...", self._mac, DEFAULT_RETRY_TIMEOUT)
+                time.sleep(DEFAULT_RETRY_TIMEOUT)
+        
+        return connected
+    
+    def _scheduleDisconnection(self):
+        '''We stablish a timer to disconnect after some time, if the user wants so'''
+        
+        if self._timer is not None:
+            self._timer.cancel()
             
-            '''We stablish a timer to disconnect after some time, if the user wants so'''
-            if timeout != NO_TIMEOUT:
-                Timer(timeout, self._disconnect).start()
-                _LOGGER.info("Auto-disconnect enabled after %d seconds.", timeout)
-        except btle.BTLEDisconnectError:
-            _LOGGER.error("Couldn't connect to Magic Switchbot device at %s", self._mac)
-            self._device = None
-            ok = False
-            # raise
-        except btle.BTLEException:
-            _LOGGER.error("Failed to connect to MagicSwitchbot.", exc_info=True)
-            self._device = None
-            ok = False
-            # raise
-            
-        return ok
+        if self._disconnectTimeout != NO_TIMEOUT:
+            self._timer = Timer(self._disconnectTimeout, self._disconnect, [True]) 
+            self._timer.start()
+            _LOGGER.info("MagicSwitchbot[%s] Auto-disconnect scheduled for %d seconds.", self._mac, self._disconnectTimeout)
 
     def _enableNotifications(self) -> bool:
         """Enable read notifications
@@ -227,35 +329,40 @@ class MagicSwitchbotDevice:
 
         notifOk = False
 
-        _LOGGER.debug("Enabling notifications for userRead characteristic (%s). Handle: 0x%X", self.UUID_USERREAD_CHAR, readHandle)
-        _LOGGER.debug("Client Characteristic Configuration Descriptor: %s. Handle: 0x%X", self.UUID_NOTIFY_SET, self._cccdDescriptor.handle)
+        _LOGGER.debug("MagicSwitchbot[%s] Enabling notifications for userRead characteristic (%s). Handle: 0x%X", self._mac, self.UUID_USERREAD_CHAR, readHandle)
+        _LOGGER.debug("MagicSwitchbot[%s] Client Characteristic Configuration Descriptor: %s. Handle: 0x%X", self._mac, self.UUID_NOTIFY_SET, self._cccdDescriptor.handle)
 
         '''Subscribe to userRed characteristic notifications'''
-        self._delegate = MagicSwitchbotDelegate(readHandle)
+        self._delegate = MagicSwitchbotDelegate(readHandle, self._mac)
         self._device.withDelegate(self._delegate)
 
         '''Enable the notifications for the read characteristic'''
         try:
             res = self._cccdDescriptor.write(binascii.a2b_hex(self.CMD_ENNOTIF), withResponse=True)
             res = self._cccdDescriptor.read()
-            _LOGGER.debug("Characteristic notifications enabled: %s", res)
+            _LOGGER.debug("MagicSwitchbot[%s] Characteristic notifications enabled: %s", self._mac, res)
             notifOk = True
         except btle.BTLEGattError as e:
-            _LOGGER.error("Error enabling notifications: %s\n", str(e))
+            _LOGGER.error("MagicSwitchbot[%s] Error enabling notifications: %s\n", self._mac, str(e))
 
         return notifOk
 
-    def _disconnect(self) -> None:
+    def _disconnect(self, scheduled=False) -> None:
         """Discconnects from the device"""
+        
+        self._timer = None
+        _LOGGER.debug("MagicSwitchbot[%s] Disconnecting%s", self._mac, " on scheduled time" if scheduled else "")
         if not self._is_connected():
+            self._device = None
+            _LOGGER.debug("MagicSwitchbot[%s] The device was not connected", self._mac)
             return
-        _LOGGER.debug("Disconnecting from MagicSwitchBot")
+        
         try:
             self._device.disconnect()
-            _LOGGER.info("Client disconnected from %s", self._mac)
+            _LOGGER.info("MagicSwitchbot[%s] Client disconnected", self._mac)
             self._token = None
         except btle.BTLEException as e:
-            _LOGGER.warning("Error disconnecting from MagicSwitchbot: %s", str(e))
+            _LOGGER.warning("MagicSwitchbot[%s] Error disconnecting: %s", self._mac, str(e))
         finally:
             self._device = None
             
@@ -267,7 +374,36 @@ class MagicSwitchbotDevice:
             bool
                 Returns True if the device is still connected
         """
-        return self._device is not None
+        
+        if self._device is not None:
+            conn_status = ""
+            
+            for i in range (self._retry_count):
+                try:
+                    '''We get the current connection state using an undocumented method from Peripheral'''
+                    if self._device is not None:
+                        conn_status = self._device.getState()
+                    else:
+                        conn_status = ""
+                except Exception as e:
+                    _LOGGER.warn("MagicSwitchbot[%s] %s when checking connection state (%d of %d attempts)", self._mac, str(e), i + 1, self._retry_count)
+                    
+                    if i < self._retry_count - 1:
+                        '''We'll retry to check the state'''
+                        time.sleep(0.5)
+                    else:
+                        '''We have an unknown state. Let's disconnect'''
+                        if self._device is not None:
+                            self._device.disconnect()
+                        self._token = None
+                        self._device = None
+            connected = (conn_status == "conn")
+        else:
+            connected = False
+            
+        _LOGGER.debug("MagicSwitchbot[%s] Connected state: %s", self._mac, "True" if connected else "False")
+        
+        return connected
             
     def _encrypt(self, data) -> str:
         """Encrypts data using AES128 ECB
@@ -281,10 +417,8 @@ class MagicSwitchbotDevice:
             str
                 Hexadecimal representation of encrypted data
         """
-        _LOGGER.debug("Unencrypted command: %s", data)
         cipher = AES.new(bytes(bytearray(self.CRYPT_KEY)), AES.MODE_ECB)
         encrypted = cipher.encrypt(bytes.fromhex(data)).hex()
-        _LOGGER.debug("Encrypted command: %s", encrypted)
         return encrypted
       
     def _decrypt(self, data) -> str:
@@ -337,7 +471,9 @@ class MagicSwitchbotDevice:
         rndLen = 32 - len(command) - len(parameter) - len(tok) - 2
         rndTail = ''.join([str(y) for _ in range(rndLen) for y in random.choice('0123456789abcdef')])
         
-        fullCommand = command + parmLen + parameter + tok + rndTail  
+        fullCommand = command + parmLen + parameter + tok + rndTail
+        
+        _LOGGER.info("MagicSwitchbot[%s] Sending %s command: %s", self._mac, self.COMMANDS[command], fullCommand)
 
         return self._encrypt(fullCommand)
 
@@ -358,18 +494,17 @@ class MagicSwitchbotDevice:
         """
       
         self._delegate.resetData()
-        _LOGGER.debug("Sending data, %s", data)
+        _LOGGER.debug("MagicSwitchbot[%s] Sending encrypted data: %s", self._mac, data)
         write_result = self._userWriteChar.write(binascii.a2b_hex(data), True)
         
         if not write_result:
-            _LOGGER.error("Sent command but didn't get a response from MagicSwitchbot confirming command was sent. "
-                          "Please check the device.")
+            _LOGGER.error("MagicSwitchbot[%s] Sent command but didn't get a response. Please check the device.", self._mac)
         else:
-            _LOGGER.info("Command sent to MagicSwitchbot (%s): %s", self._mac, data)
+            _LOGGER.debug("MagicSwitchbot[%s] Data sent OK", self._mac)
             
         return write_result
 
-    def _sendCommand(self, command, parameter, retry) -> bool:
+    def _sendCommand(self, command, parameter, retries) -> bool:
         """Sends a command to the device
         
         This method sends a command to the device via BLE, waiting and processing an execution response
@@ -380,6 +515,8 @@ class MagicSwitchbotDevice:
                 Hexadecimal string with 2 bytes for the command (and subcommand) to execute
             parameter: str
                 Hexadecimal string with 1 or more bytes as a parameter to the command
+            retries : int
+                Number of times that the connection will be retried in case of error
 
         Returns
         -------
@@ -387,10 +524,7 @@ class MagicSwitchbotDevice:
                 Returns True if the data was sent succesfully and did get a positive aknowledge after
         """
         
-        if not self._is_connected():
-            self._connect()
-            
-        if self._is_connected():
+        if self._connect(connect_timeout=self._connectTimeout, retries=retries):
             '''First of all we check if there is a token to retrieve'''
             if command != self.CMD_GETTOKEN and self._token is None:
                 '''If the command is NOT GETTOKEN, we'll issue a GETOTKEN command before sending the actual command'''
@@ -402,40 +536,49 @@ class MagicSwitchbotDevice:
                 send_success = False
                 resp_success = False
                 encrypted_command = self._prepareCommand(command, parameter)
-                
-                _LOGGER.debug("Sending command to Magicswitchbot %s", encrypted_command)
                 try:
                     send_success = self._writeData(encrypted_command)
                     if send_success:
                         ''' Wait for a response'''
                         
-                        _LOGGER.debug("Waiting for notifications from the device...")
-                        
+                        _LOGGER.debug("MagicSwitchbot[%s] Waiting for notifications...", self._mac)
+                        received = False
+                        start_time = time.time()
                         while not self._delegate.hasData():
+                            elapsed_time = time.time() - start_time
                             if self._device.waitForNotifications(1.0):
+                                received = True
                                 continue
-                            _LOGGER.debug("Waiting...")
+                            if elapsed_time >= NOTIFY_TIMEOUT:
+                                continue
+                                
+                            _LOGGER.debug("MagicSwitchbot[%s] Waiting...", self._mac)
                         
-                        encrypted_response = self._delegate.getData()
-                        _LOGGER.debug("Raw data received:  %s", encrypted_response)
-                        
-                        plain_response = self._decrypt(encrypted_response)
-                        _LOGGER.debug("Unencrypted result: %s", plain_response)
-                        resp_success = self._processResponse(plain_response)
+                        if received:
+                            encrypted_response = self._delegate.getData()
+                            _LOGGER.debug("MagicSwitchbot[%s] Encrypted data received: %s", self._mac, encrypted_response)
+                            
+                            plain_response = self._decrypt(encrypted_response)
+                            _LOGGER.debug("MagicSwitchbot[%s] Unencrypted result: %s", self._mac, plain_response)
+                            resp_success = self._processResponse(plain_response)
+                        else:
+                            _LOGGER.error("MagicSwitchbot[%s] No response received in %d seconds ", self._mac, NOTIFY_TIMEOUT)
                 except btle.BTLEException as e:
-                    _LOGGER.warning("MagicSwitchbot communication error: %s", str(e))
+                    _LOGGER.warn("MagicSwitchbot[%s] Communication error: %s", self._mac, str(e))
+                    self._disconnect()
                     
                 if resp_success:
+                    self._scheduleDisconnection()
                     return True
-                if retry < 1:
-                    _LOGGER.error("MagicSwitchbot communication failed. We won't try again.", exc_info=True)
+                if retries < 1:
+                    _LOGGER.error("MagicSwitchbot[%s] Communication failed. We won't try again.", self._mac, exc_info=True)
                     self._device = None
                     return False
                 else:
-                    _LOGGER.warning("Cannot send command to MagicSwitchbot. Retrying (remaining attempts: %d)...", retry)
+                    _LOGGER.warning("MagicSwitchbot[%s] Communication failed. Remaining attempts: %d...", self._mac, retries)
         
                 time.sleep(DEFAULT_RETRY_TIMEOUT)
-                return self._sendCommand(command, parameter, retry - 1)
+                return self._sendCommand(command, parameter, retries - 1)
             else:
                 return False
         else:
@@ -476,7 +619,7 @@ class MagicSwitchbotDevice:
         param_length = int(response[4:6])
         param = response[6:(6 + 2 * param_length)]
       
-        _LOGGER.info("Response: [Command = %s, Return Code = %s, Length = %d, Params = %s]", command, ret_code, param_length, param)
+        _LOGGER.info("MagicSwitchbot[%s] Response: (Command = %s, Return Code = %s, Length = %d, Params = %s)", self._mac, command, ret_code, param_length, param)
       
         if command == self.CMD_GETTOKEN[0:2]:
             if ret_code == self.RC_TOKENOK:
@@ -485,26 +628,26 @@ class MagicSwitchbotDevice:
                 ver_major = int(param[10:12])
                 ver_minor = int(param[12:14])
                 dev_type = param[14:16]
-                en_pwd = param[16:18]
+                en_pwd = "False" if param[16:18] == '00' else "True"
                 self._token = token 
-                _LOGGER.info("The current connection token is %s", token)
-                _LOGGER.info("Chip type: %s, Firmware version: %d.%d, Device type: %s, Password enabled: %s", chip_type, ver_major, ver_minor, dev_type, en_pwd)
+                _LOGGER.info("MagicSwitchbot[%s] The current connection token is %s", self._mac, token)
+                _LOGGER.info("MagicSwitchbot[%s] Chip type: %s, Firmware version: %d.%d, Device type: %s, Password enabled: %s", self._mac, chip_type, ver_major, ver_minor, dev_type, en_pwd)
                 success = True
             else:
-                _LOGGER.error("Error retrieving token")
+                _LOGGER.error("MagicSwitchbot[%s] Error retrieving token", self._mac)
         elif command == self.CMD_GETBAT[0:2]:
             if ret_code == self.RC_GETBAT and param.upper() != "FF":
                 self._battery = int("0x" + param, 16)
-                _LOGGER.info("Battery level retrieved: %d", self._battery)
+                _LOGGER.info("MagicSwitchbot[%s] Battery level: %d%%", self._mac, self._battery)
                 success = True
             else:
                 self._battery = None
         elif command == self.CMD_SWITCH[0:2]:
             if ret_code == self.RC_SWITCH and param == self.STA_OK:
-                _LOGGER.info("Switch state changed successfully")
+                _LOGGER.info("MagicSwitchbot[%s] Switch state changed successfully", self._mac)
                 success = True
             else:
-                _LOGGER.error("Error changing switch state")
+                _LOGGER.error("MagicSwitchbot[%s] Error changing switch state", self._mac)
                 
         return success 
 
@@ -512,15 +655,18 @@ class MagicSwitchbotDevice:
 class MagicSwitchbot(MagicSwitchbotDevice):
     """Representation of a MagicSwitchbot."""
     
-    def connect(self, timeout=NO_TIMEOUT) -> bool:
+    def connect(self, connect_timeout=DEFAULT_CONNECT_TIMEOUT, disconnect_timeout=NO_TIMEOUT) -> bool:
         """Connects to the device
         
         This method allows us to connect to the Magic Switchbot device
         
         Params
         ------
-            timeout : int
-                Specifies the ammount of time (seconds) that will be scheduled to automatically
+            connect_timeout : int (Optional)
+                Specifies the amount of time (seconds) we'll be waiting for the bluetooth device
+                to connect. If it doesn't connect on time, it returns False
+            disconnect_timeout : int (Optional)
+                Specifies the amount of time (seconds) that will be scheduled to automatically
                 disconnect from the device. If it's not specified, the client does not disconnect
                 until the object is disposed from memory
         Returns
@@ -528,7 +674,7 @@ class MagicSwitchbot(MagicSwitchbotDevice):
             bool
                 Returns True on successful connection
         """
-        return self._connect(timeout)
+        return self._connect(connect_timeout, disconnect_timeout)
     
     def auth(self) -> bool:
         """Validate the password set on the device
@@ -545,7 +691,7 @@ class MagicSwitchbot(MagicSwitchbotDevice):
     
     def disconnect(self) -> None:
         """Discconnects from the device"""
-        return self._disconnect(self)
+        return self._disconnect()
     
     def is_connected(self) -> bool:
         """Checks if the device is connected
